// SPDX-FileCopyrightText: 2024 Yury Gubich <blue@macaw.me>
// SPDX-License-Identifier: GPL-3.0-or-later

#include "logger.h"

#include <shared/definitions.h>

#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/expressions.hpp>
#include <boost/log/attributes/clock.hpp>
#include <boost/log/attributes/scoped_attribute.hpp>
#include <boost/log/support/date_time.hpp>

#include <iostream>
#include <chrono>

static constexpr std::string_view clear = "\033[0m";
static constexpr std::string_view bold = "\033[1m";

struct GlooxTag {
    gloox::LogArea bit;
    const char* label;
    const char* color;
};

static constexpr GlooxTag glooxTags[] = {
    { gloox::LogAreaClassParser,             "Parser", nullptr },
    { gloox::LogAreaClassConnectionTCPBase,  "Connection", nullptr },
    { gloox::LogAreaClassClient,             "Client", nullptr },
    { gloox::LogAreaClassClientbase,         "Client Base", nullptr },
    { gloox::LogAreaClassComponent,          "Component", nullptr },
    { gloox::LogAreaClassDns,                "DNS", nullptr },
    { gloox::LogAreaClassConnectionHTTPProxy,"HTTP Proxy", nullptr },
    { gloox::LogAreaClassConnectionSOCKS5Proxy, "SOCKS5 Proxy", nullptr },
    { gloox::LogAreaClassConnectionTCPClient,"TCP Client", nullptr },
    { gloox::LogAreaClassConnectionTCPServer,"TCP Server", nullptr },
    { gloox::LogAreaClassS5BManager,         "SOCKS5 Bytestream Manager", nullptr },
    { gloox::LogAreaClassSOCKS5Bytestream,   "SOCKS5 Bytestream", nullptr },
    { gloox::LogAreaLinkLocalManager,        "Link Local Manager", nullptr },
    { gloox::LogAreaXmlIncoming,             "XML IN", "\033[1;36m" },
    { gloox::LogAreaXmlOutgoing,             "XML OUT", "\033[1;35m" },
    { gloox::LogAreaUser,                    "USER", nullptr },
};

BOOST_LOG_ATTRIBUTE_KEYWORD(channel_attr, "Channel", std::string)
BOOST_LOG_ATTRIBUTE_KEYWORD(timestamp_attr, "TimeStamp", boost::posix_time::ptime)

Log::Logger::Logger(Level level):
    running(true),
    mutex(),
    condition(),
    thread(),
    level(level),
    sink(boost::make_shared<Sink>()),
    logger(),
    queue(8192)
{
    auto backend = sink->locked_backend();
    boost::shared_ptr<std::ostream> const stream(&std::clog, Shared::noDelete);

    backend->add_stream(stream);
    backend->auto_flush(true);

    const boost::log::filter severityFilter = boost::log::trivial::severity >= convertBoost(level);
    sink->set_filter(severityFilter);
    sink->set_formatter(&Logger::formatRecord);

    boost::log::core::get()->add_sink(sink);

    logger.add_attribute("Timestamp", boost::log::attributes::local_clock());

    thread = std::thread(&Logger::processLogQueue, this);
}

Log::Logger::~Logger() {
    running.store(false, std::memory_order_release);
    condition.notify_one();

    if (thread.joinable())
        thread.join();

    sink->stop();
    sink->flush();
    boost::log::core::get()->remove_sink(sink);
};

void Log::Logger::handleLog(gloox::LogLevel level, gloox::LogArea area, const std::string& message) {
    const Level lvl = convertGloox(level);
    if (lvl < Logger::level)
        return;

    LogEntry entry;
    entry.level = lvl;
    entry.source = LogEntry::GlooxSource(area);
    entry.message = message;

    if (!queue.push(std::move(entry)))
        std::cerr << "Log queue full, message dropped\n";
    else
        condition.notify_one();
}

std::vector<std::string> Log::Logger::convertTags(gloox::LogArea area) {
    std::vector<std::string> tags;
    for (const auto& tag : glooxTags) {
        if (area & tag.bit)
            tags.emplace_back(tag.label);
    }

    return tags;
}

std::string Log::Logger::formatDomain(const std::vector<std::string>& domain) {
    std::ostringstream tagStream;
    for (size_t i = 0; i < domain.size(); ++i) {
        if (i > 0)
            tagStream << ' ';

        tagStream << domain[i];
    }

    return tagStream.str();
}

void Log::Logger::processLogQueue() {
    LogEntry entry;
    std::unique_lock<std::mutex> lock(mutex);

    while (running.load(std::memory_order_relaxed)) {
        while (queue.pop(entry)) {
            lock.unlock();
            logEntry(entry);
            lock.lock();
        }

        condition.wait(lock, [this]() {
            return !running.load(std::memory_order_relaxed) || !queue.empty();
        });
    }


    while (queue.pop(entry))
        logEntry(entry);
}

void Log::Logger::logEntry(const LogEntry& entry) {
    std::visit(
        [this, &entry]<typename S>(const S& source) {
        if constexpr (std::is_same_v<std::decay_t<S>, LogEntry::GlooxSource>) {
            BOOST_LOG_SCOPED_LOGGER_ATTR(logger, "Channel",
                boost::log::attributes::make_constant(formatDomain(convertTags(source.area))));
            BOOST_LOG_SEV(logger, convertBoost(entry.level)) << entry.message;
        } else {
            BOOST_LOG_SCOPED_LOGGER_ATTR(logger, "Channel",
                boost::log::attributes::make_constant(formatDomain(source.domain)));
            BOOST_LOG_SEV(logger, convertBoost(entry.level)) << entry.message;
        }
    }, entry.source);
}

void Log::Logger::log(Level lvl, const std::string& message, const std::vector<std::string>& domain) const {
    if (lvl < level)
        return;

    LogEntry entry;
    entry.level = lvl;
    entry.source = LogEntry::DirectSource(domain);
    entry.message = message;

    if (!queue.push(std::move(entry)))
        std::cerr << "Log queue full, message dropped\n";
    else
        condition.notify_one();
}

void Log::Logger::formatRecord(boost::log::record_view const& rec, boost::log::formatting_ostream& out) {
    auto ts = rec[timestamp_attr];
    auto lvl = rec[boost::log::trivial::severity];
    auto msg = rec[boost::log::expressions::smessage];

    if (ts)
        out << boost::posix_time::to_simple_string(ts.get()) << " ";

    out << bold << levelColor(convertBoost(lvl.get())) << "[" << lvl.get() << "]" << clear;

    auto ch = rec[channel_attr].get_ptr();
    if (ch && !ch->empty())
        out << " [" << *ch << "]";

    out << "\t" << msg.get() << clear;
}