// SPDX-FileCopyrightText: 2024 Yury Gubich <blue@macaw.me>
// SPDX-License-Identifier: GPL-3.0-or-later

#pragma once

#include <vector>
#include <string>
#include <atomic>
#include <mutex>
#include <thread>
#include <condition_variable>
#include <variant>

#include <gloox/loghandler.h>
#include <gloox/logsink.h>

#include <boost/log/sinks/async_frontend.hpp>
#include <boost/log/sinks/text_ostream_backend.hpp>
#include <boost/lockfree/spsc_queue.hpp>

#include "log.h"

namespace Log {

class Logger: public gloox::LogHandler {
public:
    explicit Logger(Level level = Level::info);
    ~Logger() override;

    void handleLog(gloox::LogLevel level, gloox::LogArea area, const std::string& message) override;

    void log(Level level, const std::string& message, const std::vector<std::string>& domain = {}) const;
private:
    typedef boost::log::sinks::asynchronous_sink<boost::log::sinks::text_ostream_backend> Sink;
    typedef boost::log::sources::severity_logger<boost::log::trivial::severity_level> BoostLogger;

    struct LogEntry {
        struct GlooxSource {
            gloox::LogArea area;
        };

        struct DirectSource {
            std::vector<std::string> domain;
        };

        Level level;
        std::string message;
        std::chrono::system_clock::time_point timestamp;
        std::variant<GlooxSource, DirectSource> source;
    };

    std::atomic<bool> running;
    std::mutex mutex;
    mutable std::condition_variable condition;
    std::thread thread;

    Level level;
    boost::shared_ptr<Sink> sink;
    mutable BoostLogger logger;
    mutable boost::lockfree::spsc_queue<LogEntry> queue;

private:
    static void formatRecord(boost::log::record_view const& rec, boost::log::formatting_ostream& out);
    static std::vector<std::string> convertTags(gloox::LogArea area);
    static std::string formatDomain(const std::vector<std::string>& domain);
    void processLogQueue();
    void logEntry(const LogEntry& entry);
};

}